#!/usr/bin/env python3

"""NiFi worker to pass Scio produced data to the ACT platform"""

from logging import error, warning
from typing import Callable, Dict, Set, Text, cast
import argparse
import json
import os
import sys
import traceback

from act.api.helpers import handle_fact, handle_uri
from act.types.format import ValidationError, format_threat_actor, format_tool
from act.workers.libs import worker
from act.api.libs import cli
import act.api

EXTRACT_GEONAMES = ["countries"]

EXTRACT_INDICATORS = ["md5", "sha1", "sha256", "fqdn", "uri", "ipv4net"]

EXTRACT_VULNERABILITIES = ["cve", "msid"]

SCIO_GEONAMES_ACT_MAP = {
    "countries": "country",
    "regions": "region",
    "regions-derived": "region",
    "sub-regions": "subRegion",
    "sub-regions-derived": "subRegion",
}

SCIO_INDICATOR_ACT_MAP = {
    "md5": "hash",
    "sha1": "hash",
    "sha256": "hash",
    "ipv4net": "ipv4Network",
    "cve": "vulnerability",
    "msid": "vulnerability",
}

ACT_FN_MAP: Dict[Text, Callable] = {"vulnerability": lambda x: x.lower()}


def parseargs() -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = worker.parseargs("Get SCIO reports and IOCs from stdin")
    return parser


def get_scio_report() -> Dict:
    """Read scio report from stdin"""

    return cast(Dict, json.load(sys.stdin))


def report_mentions_fact(
    actapi: act.api.Act,
    object_type: Text,
    object_values: Set[Text],
    report_id: Text,
    output_format: Text,
) -> None:
    """Add mentions fact to report"""
    for value in set(object_values):
        try:
            handle_fact(
                actapi.fact("mentions")
                .source("report", report_id)
                .destination(object_type, value),
                output_format,
            )
        except ValidationError as e:
            error(str(e))
        except act.api.base.ResponseError as e:
            error("Unable to create linked fact: %s" % e)


def add_indicators_to_act(
    actapi: act.api.Act, indicators: Dict, report_id: Text, output_format: Text
) -> None:
    """Create facts of indicators map"""

    # Loop over all items under indicators in report
    for scio_indicator_type in EXTRACT_INDICATORS:
        # Get object type from ACT (default to object type in SCIO)
        act_indicator_type = SCIO_INDICATOR_ACT_MAP.get(
            scio_indicator_type, scio_indicator_type
        )

        value_fn = ACT_FN_MAP.get(act_indicator_type, lambda x: x)

        values = {value_fn(value) for value in indicators.get(scio_indicator_type, [])}

        report_mentions_fact(
            actapi, act_indicator_type, values, report_id, output_format
        )

    # For IPv4+IPv6 addresses, create mention facts
    # Use ip_obj to return exploded, normalized IPv4 and IPv6 address
    for ip in set(indicators.get("ipv4", []) + indicators.get("ipv6", [])):
        try:
            handle_fact(
                actapi.fact("mentions")
                .source("report", report_id)
                .destination(*act.api.helpers.ip_obj(ip)),
                output_format,
            )
        except ValueError as err:
            warning(f"Creating fact to {ip} fails on IP validation {err}")

    # For SHA256, create content object
    for sha256 in set(indicators.get("sha256", [])):
        handle_fact(
            actapi.fact("represents")
            .source("hash", sha256)
            .destination("content", sha256),
            output_format,
        )

    # Add emails as URI components
    for email in set(indicators.get("email", [])):
        try:
            email_uri = f"email://{email}"
            handle_uri(actapi, email_uri, output_format=output_format)

            handle_fact(
                actapi.fact("mentions")
                .source("report", report_id)
                .destination("uri", email_uri),
                output_format,
            )
        except act.api.base.ValidationError as err:
            warning(f"Fact from {email_uri} failes du to URI validation {err}")
        except act.api.schema.MissingField:
            warning(f"Unable to create facts from uri: {email_uri}")

    # Add all URI components
    for uri in set(indicators.get("uri", [])):
        try:
            handle_uri(actapi, uri, output_format=output_format)
        except act.api.base.ValidationError as err:
            warning(f"Fact from {uri} failes du to URI validation {err}")
        except act.api.schema.MissingField:
            warning(f"Unable to create facts from uri: {uri}")


def add_locations_to_act(
    actapi: act.api.Act, locations_map: Dict, report_id: Text, output_format: Text
) -> None:
    """Create facts from locations map"""

    # Locations (countries, regions, sub regions)
    for location_type in EXTRACT_GEONAMES:
        locations = {x["name"] for x in locations_map.get(location_type, [])}

        report_mentions_fact(
            actapi,
            SCIO_GEONAMES_ACT_MAP[location_type],
            locations,
            report_id,
            output_format,
        )


def add_vulnerabilities_to_act(
    actapi: act.api.Act, vulnerabilities: Dict, report_id: Text, output_format: Text
) -> None:
    """Create facts from vulnerabilities map"""

    # Locations (countries, regions, sub regions)
    for vuln_type in EXTRACT_VULNERABILITIES:
        value_fn = ACT_FN_MAP.get("vulnerability", lambda x: x)

        values = {value_fn(value) for value in vulnerabilities.get(vuln_type, [])}

        report_mentions_fact(actapi, "vulnerability", values, report_id, output_format)


def add_to_act(actapi: act.api.Act, doc: Dict, output_format: Text = "json") -> None:
    """Add a report to the ACT platform"""

    report_id: Text = doc["hexdigest"]
    filename: Text = os.path.basename(doc.get("filename", "NN"))
    title: Text = doc.get("title", filename)

    indicators: Dict = doc.get("indicators", {})
    vulnerabilities: Dict = doc.get("vulnerabilities", {})
    locations: Dict = doc.get("locations", {})

    try:
        # Report title
        handle_fact(
            actapi.fact("name", title).source("report", report_id), output_format
        )
    except act.api.base.ResponseError as e:
        error("Unable to create fact: %s" % e)

    try:
        if doc.get("uri"):
            # URI reference
            handle_fact(
                actapi.fact("represents")
                .source("report", report_id)
                .destination("content", report_id),
                output_format,
            )
            # URI reference
            handle_fact(
                actapi.fact("at")
                .source("content", report_id)
                .destination("uri", doc["uri"]),
                output_format,
            )
    except act.api.base.ResponseError as e:
        error("Unable to create fact: %s" % e)

    add_indicators_to_act(actapi, indicators, report_id, output_format)
    add_vulnerabilities_to_act(actapi, vulnerabilities, report_id, output_format)
    add_locations_to_act(actapi, locations, report_id, output_format)

    # Threat actor
    report_mentions_fact(
        actapi,
        "threatActor",
        {
            format_threat_actor(ta)
            for ta in doc.get("threatactor", {}).get("ThreatActors", [])
        },
        report_id,
        output_format,
    )

    # Tools
    report_mentions_fact(
        actapi,
        "tool",
        {format_tool(tool) for tool in doc.get("tools", {}).get("Tools", [])},
        report_id,
        output_format,
    )

    # Sector
    report_mentions_fact(
        actapi,
        "sector",
        set(doc.get("sectors", {}).get("sectors", [])),
        report_id,
        output_format,
    )


def main() -> None:
    """main function"""

    # Look for default ini file in "/etc/actworkers.ini"
    # and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = cli.handle_args(parseargs())

    actapi = worker.init_act(args)

    # Add IOCs from reports to the ACT platform
    add_to_act(
        actapi,
        get_scio_report(),
        args.output_format,
    )


def main_log_error() -> None:
    """Execute main() and log errors to error"""
    try:
        main()
    except Exception:
        error(f"Unhandled exception: {traceback.format_exc()}")
        raise


if __name__ == "__main__":
    main_log_error()



""" Tests for worker """
import sys

import _pytest
from act.workers.libs import worker
from act.api.libs import cli


def test_args_origin_name(monkeypatch: _pytest.monkeypatch.MonkeyPatch) -> None:
    """test argument origin-name"""

    origin_name = "test-origin"

    monkeypatch.setattr(sys, "argv", ["./test-worker.py", "--origin-name", origin_name])

    args = cli.handle_args(worker.parseargs("Test worker"))
    actapi = worker.init_act(args)

    assert actapi.config.origin_name == origin_name

    fact = (
        actapi.fact("mentions").source("report", "xyz").destination("fqdn", "test.com")
    )

    assert fact.origin.name == origin_name


def test_args_origin_id(monkeypatch: _pytest.monkeypatch.MonkeyPatch) -> None:
    """test argument origin-id"""

    origin_id = "00000000-0000-0000-0000-000000000001"

    monkeypatch.setattr(sys, "argv", ["./test-worker.py", "--origin-id", origin_id])

    args = cli.handle_args(worker.parseargs("Test worker"))
    actapi = worker.init_act(args)

    assert actapi.config.origin_id == origin_id

    fact = (
        actapi.fact("mentions").source("report", "xyz").destination("fqdn", "test.com")
    )

    assert fact.origin.id == origin_id



#!/usr/bin/env python3

"""hybrid-analysis.com worker for the ACT platform

Copyright 2021 the ACT project <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
"""

import argparse
import contextlib
import json
import sys
import traceback
import warnings
from functools import partialmethod
from logging import error, info
from typing import Any, Dict, Generator, List, Optional, Text

import requests

import act.api
from act.types.format import ValidationError, format_tool
from act.workers.libs import worker
from act.api.libs import cli


def parseargs() -> argparse.ArgumentParser:
    """Extract command lines argument"""

    parser = worker.parseargs("ACT hybrid-analysis.com Client")

    parser.add_argument(
        "--feed", action="store_true", help="Download the public feed only, no lookup"
    )

    parser.add_argument(
        "--apikey", default="", help="community apikey for hybrid-analysis.com"
    )

    parser.add_argument(
        "--user-agent", default="Falcon Sandbox", help="User agent while talking to API"
    )

    parser.add_argument(
        "--no-check-certificate",
        action="store_true",
        help="Do not check SSL certificate",
    )

    return parser


def download_feed(
    user_agent: Text, proxies: Optional[Dict[Text, Text]], verify_ssl: bool = True
) -> Dict[Text, Any]:
    """Download the public feed and return a dictionary"""

    url = "https://hybrid-analysis.com/feed?json"

    with ssl_verification(verify=verify_ssl):
        headers = {"User-Agent": user_agent}
        response = requests.get(url, proxies=proxies, headers=headers)

    if response.status_code != 200:
        raise CommunicationError(
            f"hybrid_analysis_feed.download_feed() could not download public feed, "
            f"error calling {url}: Status = {response.status_code}"
        )

    try:
        data: Dict[Text, Any] = response.json()
    except json.decoder.JSONDecodeError as err:
        raise CommunicationError(
            f"hybrid_analysis_feed.download_feed() could not load public feed, "
            f"error decoding json result from {url}: {err}"
        )

    return data


def handle_feed(
    actapi: act.api.Act,
    user_agent: Text,
    proxies: Optional[Dict[Text, Text]] = None,
    verify_ssl: bool = True,
    output_format: Text = "json",
) -> None:
    """Download, parse and provide facts from the public feed of hybrid-analysis.com"""

    feed = download_feed(user_agent, proxies, verify_ssl)

    feeds_facts: List[act.api.fact.Fact] = []

    for report in feed["data"]:
        if not (report.get("isinteresting", False) or report.get("threatlevel", 0)):
            continue
        # store data if threatlevel > 0 or report is interesting

        if "sha256" not in report:
            continue

        feeds_facts += handle_report(actapi, report)

    for fact in feeds_facts:
        act.api.helpers.handle_fact(fact, output_format=output_format)


def handle_hosts(
    actapi: act.api.Act, content: Text, hosts: List[Text]
) -> List[act.api.fact.Fact]:
    """handle the hosts part of a hybrid-analysis report"""

    feeds_facts: List[act.api.fact.Fact] = []

    for host in hosts:
        (ip_type, ip) = act.api.helpers.ip_obj(host)

        chain = []

        chain.append(
            actapi.fact("connectsTo").source("content", content).destination("uri", "*")
        )
        chain.append(
            actapi.fact("resolvesTo").source("fqdn", "*").destination(ip_type, ip)
        )
        chain.append(
            actapi.fact("componentOf").source("fqdn", "*").destination("uri", "*")
        )

        feeds_facts += act.api.fact.fact_chain(*chain)

    return feeds_facts


def handle_domains(
    actapi: act.api.Act, content: Text, domains: List[Text]
) -> List[act.api.fact.Fact]:
    """Handle the domains part of a hybrid-analysis report"""

    feeds_facts: List[act.api.fact.Fact] = []

    for domain in domains:

        chain = []

        chain.append(
            actapi.fact("connectsTo").source("content", content).destination("uri", "*")
        )
        chain.append(
            actapi.fact("componentOf").source("fqdn", domain).destination("uri", "*")
        )

        feeds_facts += act.api.fact.fact_chain(*chain)

    return feeds_facts


def handle_extracted_files(
    actapi: act.api.Act, content: Text, extracted_files: List[Dict]
) -> List[act.api.fact.Fact]:
    """Handle the extracted_files part of a hybrid_analysis report"""

    feeds_facts: List[act.api.fact.Fact] = []

    for file in extracted_files:

        chain = []

        if "sha256" not in file:
            continue

        if not file["file_path"]:
            info(f"{file} is missing file_path using name instead")

        path = file["file_path"] if file["file_path"] else file["name"]

        chain.append(
            actapi.fact("componentOf").source("path", path).destination("uri", "*")
        )

        chain.append(
            actapi.fact("at").source("content", file["sha256"]).destination("uri", "*")
        )

        feeds_facts += act.api.fact.fact_chain(*chain)

        for hash_type in ["md5", "sha1", "sha256"]:
            feeds_facts.append(
                actapi.fact("represents")
                .source("hash", file[hash_type])
                .destination("content", file["sha256"])
            )
            feeds_facts.append(
                actapi.fact("category", hash_type).source("hash", file[hash_type])
            )

        if (
            content != file["sha256"]
        ):  # the act platform does not accept same object on source and destination for write
            feeds_facts.append(
                actapi.fact("writes")
                .source("content", content)
                .destination("content", file["sha256"])
            )

    return feeds_facts


def handle_classification_tags(
    actapi: act.api.Act, content: Text, classification_tags: List[Text]
) -> List[act.api.fact.Fact]:
    """handle the classification_tags part or a hybrid_analysis report"""

    feeds_facts: List[act.api.fact.Fact] = []

    for tag in classification_tags:
        try:
            feeds_facts.append(
                actapi.fact("classifiedAs")
                .source("content", content)
                .destination("tool", format_tool(tag))
            )
        except ValidationError as e:
            error(str(e))

    return feeds_facts


def handle_mitre_attcks(
    actapi: act.api.Act, content: Text, mitre_attcks: List[Dict]
) -> List[act.api.fact.Fact]:
    """Handle the MITRE Att&ck part of the hybrid analysis report"""

    feeds_facts: List[act.api.fact.Fact] = []

    for attck in mitre_attcks:
        chain = []

        chain.append(
            actapi.fact("classifiedAs")
            .source("content", content)
            .destination("tool", "*")
        )
        chain.append(
            actapi.fact("implements")
            .source("tool", "*")
            .destination("technique", attck["technique"])
        )

        feeds_facts += act.api.fact.fact_chain(*chain)

    return feeds_facts


def handle_process_list(
    actapi: act.api.Act, content: Text, process_list: List[Dict]
) -> List[act.api.fact.Fact]:
    """Handle the process list part of the hybrid analysis report"""

    feeds_facts: List[act.api.fact.Fact] = []

    for proc in process_list:

        chain = []

        path = proc["normalizedpath"] if "normalizedpath" in proc else proc["name"]

        chain.append(
            actapi.fact("executes").source("content", content).destination("uri", "*")
        )
        chain.append(
            actapi.fact("componentOf").source("path", path).destination("uri", "*")
        )

        feeds_facts += act.api.fact.fact_chain(*chain)

    return feeds_facts


def handle_report(
    actapi: act.api.Act, report: Dict[Text, Any]
) -> List[act.api.fact.Fact]:
    """Create facts from a report"""

    feeds_facts: List[act.api.fact.Fact] = []

    content = report["sha256"]
    for hash_type in ["md5", "sha1", "sha256", "ssdeep", "imphash", "sha512"]:
        if (
            hash_type not in report
            or not report[hash_type]
            or report[hash_type] == "Unknown"
        ):
            info(f"{hash_type} not set for content {content}")
            continue
        feeds_facts.append(
            actapi.fact("represents")
            .source("hash", report[hash_type])
            .destination("content", content)
        )
        feeds_facts.append(
            actapi.fact("category", hash_type).source("hash", report[hash_type])
        )

    feeds_facts += handle_hosts(actapi, content, report.get("hosts", []))
    feeds_facts += handle_domains(actapi, content, report.get("domains", []))
    feeds_facts += handle_extracted_files(
        actapi, content, report.get("extracted_files", [])
    )
    feeds_facts += handle_classification_tags(
        actapi, content, report.get("classification_tags", [])
    )
    feeds_facts += handle_mitre_attcks(actapi, content, report.get("mitre_attcks", []))
    feeds_facts += handle_process_list(actapi, content, report.get("process_list", []))

    return feeds_facts


def handle_hash(
    actapi: act.api.Act,
    apikey: Text,
    hashdigest: Text,
    user_agent: Text,
    proxies: Optional[Dict[Text, Text]] = None,
    verify_ssl: bool = True,
    output_format: Text = "json",
) -> None:
    """Download, parse and provide facts from the public feed of hybrid-analysis.com"""

    data = search_hash(apikey, hashdigest, user_agent, proxies, verify_ssl)

    for report in data:
        for fact in handle_report(actapi, report):
            act.api.helpers.handle_fact(fact, output_format=output_format)


def search_hash(
    apikey: Text,
    hashdigest: Text,
    user_agent: Text,
    proxies: Optional[Dict[Text, Text]] = None,
    verify_ssl: bool = True,
) -> List[Dict[Text, Any]]:
    """Search the hybrid-analysis api for a specific hash"""

    url = "https://www.hybrid-analysis.com/api/v2/search/hash"

    with ssl_verification(verify=verify_ssl):
        headers = {
            "User-Agent": user_agent,
            "accept": "application/json",
            "api-key": apikey,
            "Content-Type": "application/x-www-form-urlencoded",
        }

        form_data = {"hash": hashdigest}

        response = requests.post(url, proxies=proxies, headers=headers, data=form_data)

    if response.status_code != 200:
        print(response.text)
        raise CommunicationError(
            f"hybrid_analysis_feed.search_hash() could not search community API, "
            f"error calling {url}: Status = {response.status_code}"
        )

    try:
        data: List[Dict[Text, Any]] = response.json()
    except json.decoder.JSONDecodeError as err:
        raise CommunicationError(
            f"hybrid_analysis_feed.search_hash() could not load search result, "
            f"error decoding json result from {url}: {err}"
        )

    return data


def main() -> None:
    """main function"""

    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = cli.handle_args(parseargs())

    actapi = worker.init_act(args)

    # if not args.apikey:
    #    cli.fatal("You must specify --apikey on command line or in config file")

    proxies = (
        {"http": args.proxy_string, "https": args.proxy_string}
        if args.proxy_string
        else None
    )

    params = {
        "actapi": actapi,
        "user_agent": args.user_agent,
        "proxies": proxies,
        "verify_ssl": args.no_check_certificate,
        "output_format": args.output_format,
    }

    if args.feed:
        handle_feed(**params)
    else:
        params["apikey"] = args.apikey
        for line in sys.stdin:
            params["hashdigest"] = line.strip()
            handle_hash(**params)


@contextlib.contextmanager
def ssl_verification(verify: bool = True) -> Generator[None, None, None]:
    """Monkey patch request to manage ssl verification. Can be used 'around' code
    that uses requests internally"""

    old_request = requests.Session.request
    requests.Session.request = partialmethod(old_request, verify=verify)  # type: ignore

    warnings.filterwarnings("ignore", "Unverified HTTPS request")
    yield
    warnings.resetwarnings()

    requests.Session.request = old_request  # type: ignore


class CommunicationError(Exception):
    """CommunicationError is used to gather all communication errors into one"""

    ...


def main_log_error() -> None:
    """Main entry point, catching and logging errors"""

    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()



#!/usr/bin/env python3

"""ip filter plugin


Copyright 2020 mnemonic AS <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

echo the first line and return exit status based on IP type:

Filter out all lines if parsed as IP address and IP is

- Multicast (RFC 3171, RFC 2373)
- Private (RC 1918, RFC 4193)
- Unspecified (RFC 5735, RFC 2373)
- Reserved (IETF)
- Loopback (RF C3330, RFC 2372)
- Link local (RF3927)

"""


import ipaddress
import sys
import traceback
from logging import debug, error

from act.workers.libs import worker
from act.api.libs import cli


def process() -> None:
    """Read ip addresses from stdin"""

    output_rows = 0

    for line in sys.stdin:
        line = line.strip()

        if not line:
            continue

        try:
            ip = ipaddress.ip_address(line)
        except ValueError:
            # No IP Address
            print(line)
            output_rows += 1
            continue

        if ip.is_multicast:
            #  reserved for multicast use. See RFC 3171 (for IPv4) or RFC 2373 (for IPv6).
            continue

        if ip.is_private:
            # private networks. See RFC 1918 (for IPv4) or RFC 4193 (for IPv6).
            continue

        if ip.is_unspecified:
            # unspecified. See RFC 5735 (for IPv4) or RFC 2373 (for IPv6).
            continue

        if ip.is_reserved:
            # IETF reserved.
            continue

        if ip.is_loopback:
            # loopback address. See RFC 3330 (for IPv4) or RFC 2373 (for IPv6).
            continue

        if ip.is_link_local:
            # link-local usage. See RFC 3927.
            continue

        print(line)
        output_rows += 1

    if not output_rows:
        debug("All rows are filtered out")


def main() -> None:
    """Main function"""
    args = cli.handle_args(worker.parseargs("IP Filter"))
    worker.init_act(args)
    process()


def main_log_error() -> None:
    "Main function wrapper. Log all exceptions to error"
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()



#!/usr/bin/env python3

"""
Thaicert worker for the ACT platform

"""

import argparse
import traceback
import act
import act.api

from collections import defaultdict
from typing import Text, List
from itertools import combinations
from logging import error, warning

from act.types.format import ValidationError, format_threat_actor, format_tool
from act.workers.libs import worker
from act.api.helpers import handle_fact
from act.api.libs import cli


WORKER_NAME = "thaicert"
THAICERT_URL = "https://apt.thaicert.or.th/cgi-bin/getcard.cgi?g=all&o=j"
STIX_VOCAB = "http://raw.githubusercontent.com/oasis-open/cti-stix2-json-schemas/stix2.1/schemas/sdos/identity.json"
COUNTRY_REGIONS = "https://raw.githubusercontent.com/lukes/ISO-3166-Countries-with-Regional-Codes/master/all/all.json"
THAICERT_TOOLS_URL = "https://apt.thaicert.or.th/cgi-bin/getcard.cgi?t=all&o=j"


def parseargs() -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = worker.parseargs("ThaiCERT worker")
    parser.add_argument(
        "--url", dest="thaicert_url", default=THAICERT_URL, help="ThaiCERT data URL"
    )
    return parser


def process(client: act.api.Act, ta_cards: List, output_format: Text = "json") -> None:
    "Extract threat actor cards from ThaiCERT"

    # Keep list of names with an alias points to
    # There should only be one to have a meaningfull alias
    ta_alias_names = defaultdict(set)
    ta_aliases = {}

    for actor in ta_cards:
        try:
            ta_name = format_threat_actor(actor["names"][0]["name"])
        except ValidationError as e:
            error(str(e))
            continue

        ta_aliases[ta_name] = []

        for alias in actor["names"][1:]:
            try:
                alias_name = format_threat_actor(alias["name"])
            except ValidationError as e:
                error(str(e))
                continue

            # names might be identical after format/normalization
            if ta_name == alias_name:
                continue

            ta_alias_names[alias_name].add(ta_name)
            ta_aliases[ta_name].append(alias_name)

    for ta_name in ta_aliases:
        for alias_name in ta_aliases[ta_name]:

            # This alias is mentioned for multiple main threat
            # actors so we skip this alias
            if len(ta_alias_names[alias_name]) > 1:
                warning(
                    f"Skipping TA alias {alias_name} <-> {ta_name} since {alias_name} is alias for multiple names: {ta_alias_names[alias_name]}"
                )
                continue

            handle_fact(
                client.fact("alias").bidirectional(
                    "threatActor", ta_name, "threatActor", alias_name
                ),
                output_format=output_format,
            )

    for actor in ta_cards:
        if "operations" in actor:
            for operation in actor["operations"]:
                if len(operation["activity"].split("\n")[0].split()) < 4:
                    for ta in actor["actor"].split(","):
                        try:
                            chain = act.api.fact.fact_chain(
                                client.fact("attributedTo")
                                .source("incident", "*")
                                .destination(
                                    "campaign", operation["activity"].split("\n")[0]
                                ),
                                client.fact("attributedTo")
                                .source("incident", "*")
                                .destination("threatActor", format_threat_actor(ta)),
                            )
                        except ValidationError as e:
                            error(str(e))
                            continue

                        for fact in chain:
                            try:
                                handle_fact(fact, output_format=output_format)
                            except act.api.base.ValidationError as err:
                                warning(
                                    "ValidationError while storing objects: %s" % err
                                )


def add_countries(
    client: act.api.Act, ta_cards: List, countries: List, output_format: Text = "json"
) -> None:
    """
    Only submit country if ISO-3166 country
    """
    for actor in ta_cards:
        if "country" in actor:
            for country in actor["country"]:
                if country.lower() in countries:
                    try:
                        chain = act.api.fact.fact_chain(
                            client.fact("locatedIn")
                            .source("organization", "*")
                            .destination("country", country),
                            client.fact("attributedTo")
                            .source("threatActor", format_threat_actor(actor["actor"]))
                            .destination("organization", "*"),
                        )
                    except ValidationError as e:
                        error(str(e))
                        continue

                    for fact in chain:
                        handle_fact(fact, output_format=output_format)

        if "observed-countries" in actor:
            for country in actor["observed-countries"]:
                if country.lower() in countries:
                    for ta in actor["actor"].split(","):
                        try:
                            chain = act.api.fact.fact_chain(
                                client.fact("locatedIn")
                                .source("organization", "*")
                                .destination("country", country),
                                client.fact("targets")
                                .source("incident", "*")
                                .destination("organization", "*"),
                                client.fact("attributedTo")
                                .source("incident", "*")
                                .destination("threatActor", format_threat_actor(ta)),
                            )
                        except ValidationError as e:
                            error(str(e))
                            continue

                        for fact in chain:
                            handle_fact(fact, output_format=output_format)


def add_sectors(
    client: act.api.Act, ta_cards: List, vocab: List, output_format: Text = "json"
) -> None:
    """
    Only submit sectors if in STIX vocabulary
    """
    for actor in ta_cards:
        if "observed-sectors" in actor:
            for sector in actor["observed-sectors"]:
                if sector.lower() in vocab["definitions"]["industry-sector-ov"]["enum"]:
                    for ta in actor["actor"].split(","):
                        try:
                            chain = act.api.fact.fact_chain(
                                client.fact("memberOf")
                                .source("organization", "*")
                                .destination("sector", sector.lower()),
                                client.fact("targets")
                                .source("incident", "*")
                                .destination("organization", "*"),
                                client.fact("attributedTo")
                                .source("incident", "*")
                                .destination("threatActor", format_threat_actor(ta)),
                            )
                        except ValidationError as e:
                            error(str(e))
                            continue

                        for fact in chain:
                            handle_fact(fact, output_format=output_format)


def add_tools(
    client: act.api.Act, ta_cards: List, tools: List, output_format: Text = "json"
) -> None:
    """
    Submit tool aliases and actor tools
    """
    tool_vocab = [tool["tool"] for tool in tools]
    for actor in ta_cards:
        if "tools" in actor:
            for tool in actor["tools"]:
                if tool in tool_vocab:
                    for ta in actor["actor"].split(","):
                        try:
                            chain = act.api.fact.fact_chain(
                                client.fact("classifiedAs")
                                .source("content", "*")
                                .destination("tool", format_tool(tool)),
                                client.fact("observedIn")
                                .source("content", "*")
                                .destination("incident", "*"),
                                client.fact("attributedTo")
                                .source("incident", "*")
                                .destination("threatActor", format_threat_actor(ta)),
                            )
                        except ValidationError as e:
                            error(str(e))
                            continue

                        for fact in chain:
                            try:
                                handle_fact(fact, output_format=output_format)
                            except act.api.base.ValidationError as err:
                                error("ResponseError while storing objects: %s" % err)

    for values in tools:
        aliases = set(
            [tool["name"].strip() for tool in values["names"]] + [values["tool"]]
        )
        for tool1, tool2 in combinations(aliases, 2):
            try:
                tool1 = format_tool(tool1)
                tool2 = format_tool(tool2)
            except ValidationError as e:
                error(str(e))
                continue

            if tool1 == tool2:
                continue

            fact = client.fact("alias").bidirectional("tool", tool1, "tool", tool2)
            handle_fact(fact)


def main() -> None:
    """Main function"""
    args = cli.handle_args(parseargs())
    actapi = worker.init_act(args)
    ta_cards = worker.fetch_json(
        args.thaicert_url, args.proxy_string, args.http_timeout
    )
    process(actapi, ta_cards["values"])

    vocab = worker.fetch_json(STIX_VOCAB, args.proxy_string, args.http_timeout)
    add_sectors(actapi, ta_cards["values"], vocab)

    countries = worker.fetch_json(COUNTRY_REGIONS, args.proxy_string, args.http_timeout)
    countries = [country["name"].lower() for country in countries]
    add_countries(actapi, ta_cards["values"], countries)

    tools = worker.fetch_json(THAICERT_TOOLS_URL, args.proxy_string, args.http_timeout)
    add_tools(actapi, ta_cards["values"], tools["values"])


def main_log_error() -> None:
    "Main function wrapper. Log all exceptions to error"
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()



#!/usr/bin/env python3
import json
import os
import argparse
from act.workers.libs.ta_list_methods import (
    get_all_ta_from_act,
    get_all_alias_facts_from_act,
    add_ta_to_map,
    add_ta_alias_to_map,
    create_config,
)

from act.workers.libs import worker
from act.api.libs import cli


def parseargs() -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = worker.parseargs(
        "Updates list of threat actor aliases from an ACT platform"
    )
    parser.add_argument(
        "--aliasfile",
        required=True,
        help=".cfg-file with \
        existing threat actor aliases",
    )
    parser.add_argument(
        "--newaliasfile",
        required=True,
        help="name of \
        .cfg-file with updated list of threat actor aliases",
    )
    parser.add_argument(
        "--output-json",
        action="store_true",
        help="Enable \
        this flag if json testfiles should be created.",
    )

    return parser


def main():
    """main function"""

    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = cli.handle_args(parseargs())

    # gets all ta names from objects(as a set of strings) and facts(as a
    # set with tuples of two strings) in ACT.

    threatactors = get_all_ta_from_act(args.baseurl, args.userid)

    # gets all threat actor aliases from act platform

    ta_aliases = get_all_alias_facts_from_act(args.baseurl, args.userid)

    # save ta and ta_aliases to json test file
    if args.output_json:
        with open("test/objects.json", "w") as outfile:
            outfile.write(json.dumps(list(threatactors)))

        with open("test/facts.json", "w") as outfile:
            outfile.write(json.dumps(list(ta_aliases)))

    # adds all ta names from threatActor objects from ACT into ta_map
    ta_map = add_ta_to_map(threatactors)

    # adds all ta names from alias-facts in ACT to the ta_map
    ta_map_with_aliases = add_ta_alias_to_map(ta_aliases, ta_map)

    # creates a new .cfg-file including all the content from ta_map
    # and the existing cfg.-file from arguments.
    create_config(ta_map_with_aliases, args.aliasfile, args.newaliasfile)


if __name__ == "__main__":

    main()



""" Test for argus case worker """
import json

import act.api

from act.workers.libs import argus


def test_refang() -> None:
    """refang tests"""

    assert argus.refang_uri("hxxp://www[.]mnemonic[.]no") == "http://www.mnemonic.no"
    assert (
        argus.refang_uri("hxxp://www.mnemonic.no/hxxp") == "http://www.mnemonic.no/hxxp"
    )
    assert argus.refang_uri("hXXps://www(.)mnemonic(.)no") == "https://www.mnemonic.no"
    assert (
        argus.refang_uri("hxxp://www[.]mnemonic[.]no/abc%2fdef")
        == "http://www.mnemonic.no/abc/def"
    )


# pylint: disable=too-many-locals


def test_argus_case_facts(capsys, caplog) -> None:  # type: ignore
    """Test for argus case facts, by comparing to captue of stdout"""
    with open("test/data/argus-event.json") as argus_event:
        event = json.loads(argus_event.read())

    api = act.api.Act("", None, "error")
    act.api.helpers.handle_fact.cache_clear()

    argus.handle_argus_event(
        api,
        event,
        content_props=["file.sha256", "process.sha256"],
        hash_props=[
            "file.md5",
            "process.md5",
            "file.sha1",
            "process.sha1",
            "file.sha512",
            "process.sha512",
        ],
        output_format="str",
    )

    captured = capsys.readouterr()
    facts = set(captured.out.split("\n"))
    logs = [rec.message for rec in caplog.records]

    print(captured.out)

    prop = event["properties"]
    uri1 = event["uri"]
    uri2 = "http://test-domain2.com/path.cgi"
    uri3 = "http://test-domain3.com/abc"
    case_id = "ARGUS-{}".format(event["associatedCase"]["id"])
    observationTime = event["startTimestamp"]

    signature = event["attackInfo"]["signature"]

    # Fact chain from md5 hash through content to incident
    md5_chain = act.api.fact.fact_chain(
        api.fact("represents")
        .source("hash", prop["file.md5"])
        .destination("content", "*"),
        api.fact("observedIn").source("content", "*").destination("incident", case_id),
    )

    # Fact chain from event through technique to tactic
    tactic_chain = act.api.fact.fact_chain(
        api.fact("observedIn")
        .source("technique", "*")
        .destination("incident", case_id),
        api.fact("implements")
        .source("technique", "*")
        .destination("tactic", "Discovery"),
    )

    sha256 = "01ba4719c80b6fe911b091a7c05124b64eeece964e09c058ef8f9805daca546b"

    fact_assertions = (
        api.fact("observedIn")
        .source("content", sha256)
        .destination("incident", case_id),
        api.fact("name", "Infected host").source("incident", case_id),
        api.fact("observedIn").source("uri", uri1).destination("incident", case_id),
        api.fact("observedIn").source("uri", uri2).destination("incident", case_id),
        api.fact("observedIn").source("uri", uri3).destination("incident", case_id),
        api.fact("componentOf")
        .source("fqdn", "test-domain.com")
        .destination("uri", uri1),
        api.fact("componentOf").source("path", "/path.cgi").destination("uri", uri1),
        api.fact("scheme", "http").source("uri", uri1),
        api.fact("observedIn")
        .source("uri", "tcp://1.2.3.4")
        .destination("incident", case_id),
    )

    # All facts should have a corresponding meta fact observationTime
    meta_fact_assertions = [
        fact.meta("observationTime", str(observationTime))
        for fact in fact_assertions + tactic_chain + md5_chain
    ]

    fact_negative_assertions = [
        # signature is removed from the data model in 2.0
        api.fact("detects")
        .source("signature", signature)
        .destination("incident", case_id),
        # This fact should not exist, since we only add IPs with public addresses
        api.fact("observedIn")
        .source("uri", "tcp://192.168.1.1")
        .destination("incident", case_id),
        # This fact should not exist, since it does not have scheme
        api.fact("observedIn")
        .source("uri", "illegal-url.com")
        .destination("incident", case_id),
        # We have URI, so this should not be constructed from the fqdn
        api.fact("observedIn")
        .source("uri", "tcp://test-domain.com")
        .destination("incident", case_id),
        # Not valid content hash (sha256)
        api.fact("observedIn")
        .source("content", "bogus")
        .destination("incident", case_id),
    ]

    assert 'Illegal sha256: "bogus" in property "file.sha256"' in logs

    for fact_assertion in fact_assertions:
        assert str(fact_assertion) in facts

    for fact_assertion in fact_negative_assertions:
        assert str(fact_assertion) not in facts

    for fact_assertion in md5_chain:
        assert str(fact_assertion) in facts

    for fact_assertion in tactic_chain:
        assert str(fact_assertion) in facts

    for fact_assertion in meta_fact_assertions:
        assert str(fact_assertion) in facts



#!/usr/bin/env python3
""" Handle worker config"""

import argparse
import os
import sys
from typing import List, Text

import caep
from pkg_resources import resource_string

from act.api.libs import cli


def string_list(data: Text) -> List[Text]:
    """Split the argument into list of elements based on , separator"""

    if not data:
        return list()

    return [element.strip() for element in data.split(",")]


def parseargs() -> argparse.Namespace:
    """Parse arguments"""

    parser = argparse.ArgumentParser(
        "ACT worker config",
        epilog="""
    show - Print default config

    user - Copy default config to {0}/{1}

    system - Copy default config to /etc/{1}

""".format(
            caep.get_config_dir(cli.CONFIG_ID), cli.CONFIG_NAME
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument("action", nargs=1, choices=["show", "user", "system"])

    return parser.parse_args()


def default_ini() -> Text:
    "Get content of default ini file"
    return resource_string("act.workers", "etc/{}".format(cli.CONFIG_NAME)).decode(
        "utf-8"
    )


def save_config(filename: Text) -> None:
    """Save config to specified filename"""
    if os.path.isfile(filename):
        sys.stderr.write("Config already exists: {}\n".format(filename))
        sys.exit(1)

    try:
        with open(filename, "w") as f:
            f.write(default_ini())
    except PermissionError as err:
        sys.stderr.write("{}\n".format(err))
        sys.exit(2)

    print("Config copied to {}".format(filename))


def main() -> None:
    "main function"
    args = parseargs()

    if "show" in args.action:
        print(default_ini())

    if "user" in args.action:
        config_dir = caep.get_config_dir(cli.CONFIG_ID, create=True)
        save_config(os.path.join(config_dir, cli.CONFIG_NAME))

    if "system" in args.action:
        save_config("/etc/{}".format(cli.CONFIG_NAME))


if __name__ == "__main__":
    main()



""" Argus module to create facts from argus events """

import re
from collections import defaultdict
from ipaddress import AddressValueError, IPv4Address
from logging import debug, error, info, warning
from typing import Dict, List, Optional, Text, cast

import act.api
import act.api.helpers

HASH_MD5_RE = re.compile(r"^[0-9a-f]{32}$")
HASH_SHA1_RE = re.compile(r"^[0-9a-f]{40}$")
HASH_SHA256_RE = re.compile(r"^[0-9a-f]{64}$")
HASH_SHA5112_RE = re.compile(r"^[0-9a-f]{128}$")


def is_public_ip(ip_str: Text) -> bool:
    """Return True if IP (str) is a valid, public IP"""
    try:
        return IPv4Address(ip_str).is_global
    except AddressValueError:
        return False


def refang_uri(uri: Text) -> Text:
    """Return refanged uri, e.g. hxxp://www[.]mnemonic[.]no -> http://www.mnemonic.no"""

    return (
        re.sub(r"^hxxp", "http", uri, flags=re.I)
        .replace("[.]", ".")
        .replace("(.)", ".")
        .replace("{.}", ".")
        .replace("%2f", "/")
    )


def handle_fact_wrapper(
    fact: act.api.fact.Fact, output_format: Text, observationTime: Optional[int] = None
) -> None:
    """wrap act.helpers.handle_fact and log all errors (and continue)"""
    try:
        f = act.api.helpers.handle_fact(fact, output_format)

        if f and observationTime:
            act.api.helpers.handle_fact(
                f.meta("observationTime", str(observationTime)), output_format
            )

    except act.api.base.ResponseError:
        error("Error adding fact (ResponseError): {}".format(fact, exc_info=True))
    except act.api.schema.MissingField:
        error("Error adding fact (missing field): {}".format(fact, exc_info=True))


def handle_uri_wrapper(
    actapi: act.api.Act,
    uri: Text,
    output_format: Text,
    observationTime: Optional[int] = None,
) -> bool:
    """wrap act.helpers.handle_uri and log all errors. Return True on success and False on failure"""
    try:

        for f in act.api.helpers.handle_uri(actapi, uri, output_format):

            if observationTime:
                act.api.helpers.handle_fact(
                    f.meta("observationTime", str(observationTime)), output_format
                )

    except act.api.base.ValidationError:
        info("Error adding uri (ValidationError): {}".format(uri, exc_info=True))
        return False
    except act.api.base.ResponseError:
        error("Error adding uri (ResponseError): {}".format(uri, exc_info=True))
        return False
    except act.api.schema.MissingField:
        warning("Error adding uri (missing field): {}".format(uri, exc_info=True))
        return False

    return True


def handle_argus_event_technique(
    actapi: act.api.Act,
    properties: defaultdict,
    case_id: Text,
    output_format: Text,
    observationTime: Optional[int] = None,
) -> None:
    """Handle MITRE ATT&CK technique in events"""

    for technique in properties["mitreAttack.techniqueId"].split("\n"):
        handle_fact_wrapper(
            actapi.fact("observedIn")
            .source("technique", technique.strip())
            .destination("incident", case_id),
            output_format=output_format,
            observationTime=observationTime,
        )


def handle_argus_event_sub_technique(
    actapi: act.api.Act,
    properties: defaultdict,
    case_id: Text,
    output_format: Text,
    observationTime: Optional[int] = None,
) -> None:
    """Handle MITRE ATT&CK sub technique in events"""

    for sub_technique in properties["mitreAttack.subTechniqueId"].split("\n"):
        handle_fact_wrapper(
            actapi.fact("observedIn")
            .source("technique", sub_technique.strip())
            .destination("incident", case_id),
            output_format=output_format,
            observationTime=observationTime,
        )


def handle_argus_event_tactic(
    actapi: act.api.Act,
    properties: defaultdict,
    case_id: Text,
    output_format: Text,
    observationTime: Optional[int] = None,
) -> None:
    """Handle MITRE ATT&CK tactic in events"""

    for tactic in properties["mitreAttack.tactic"].split("\n"):
        chain = act.api.fact.fact_chain(
            actapi.fact("observedIn")
            .source("technique", "*")
            .destination("incident", case_id),
            actapi.fact("implements")
            .source("technique", "*")
            .destination("tactic", tactic.strip()),
        )

        for fact in chain:
            handle_fact_wrapper(
                fact, output_format=output_format, observationTime=observationTime
            )


def handle_argus_event_hash(
    actapi: act.api.Act,
    properties: defaultdict,
    case_id: Text,
    content_props: List[Text],
    hash_props: List[Text],
    output_format: Text,
    observationTime: Optional[int] = None,
) -> None:
    """Handle file/process hash properties in events"""

    # For all hash values that represents a content object (typically sha256)
    for prop in content_props:
        if properties[prop]:
            # Properties can be aggregated (separated by newlines)
            for value in properties[prop].split("\n"):
                value = value.lower()
                if not HASH_SHA256_RE.search(value):
                    warning('Illegal sha256: "{}" in property "{}"'.format(value, prop))
                    continue

                handle_fact_wrapper(
                    actapi.fact("observedIn")
                    .source("content", value)
                    .destination("incident", case_id),
                    output_format=output_format,
                    observationTime=observationTime,
                )

                # A SHA256 hash is represented by itself
                handle_fact_wrapper(
                    actapi.fact("represents")
                    .source("hash", value)
                    .destination("content", value),
                    output_format=output_format,
                    observationTime=observationTime,
                )

    # For all hash values that needs a placeholder (unknown content/sha256)
    for prop in hash_props:
        if properties[prop]:
            for value in properties[prop].split("\n"):
                value = value.lower()
                if not (
                    HASH_MD5_RE.search(value)
                    or HASH_SHA1_RE.search(value)
                    or HASH_SHA5112_RE.search(value)
                ):
                    warning("Unknown hash: {} in property {}".format(value, prop))
                    continue

                if value:
                    chain = act.api.fact.fact_chain(
                        actapi.fact("represents")
                        .source("hash", value)
                        .destination("content", "*"),
                        actapi.fact("observedIn")
                        .source("content", "*")
                        .destination("incident", case_id),
                    )

                    for fact in chain:
                        handle_fact_wrapper(
                            fact,
                            output_format=output_format,
                            observationTime=observationTime,
                        )


def get_scheme(event: defaultdict) -> Text:
    "Get scheme based on protocol, default to 'network'"

    protocols = ("tcp", "udp", "icmp")

    # If we have a known protocol, use this as scheme when constructing URI
    if event["protocol"] in protocols:
        return cast(Text, event["protocol"])
    return "network"


def handle_argus_event_ip(
    actapi: act.api.Act,
    event: defaultdict,
    case_id: Text,
    output_format: Text,
    observationTime: Optional[int] = None,
) -> None:
    "Create facts from source/destination IP address"

    scheme = get_scheme(event)

    for direction in ["source", "destination"]:
        # ensure we have IP information for this direction
        if not event[direction] and event[direction]["networkAddress"]:
            continue

        address = event[direction]["networkAddress"]["address"]

        if not is_public_ip(address):
            continue

        # Check if the flag <direction>_IS_CUSTOMERNET is set
        if "{}_IS_CUSTOMERNET".format(direction.upper()) in event["flags"]:
            debug("Address is customernet: {}, event={}".format(address, event))
            continue

        uri = "{}://{}".format(scheme, address)

        # Facts: uri components
        if not handle_uri_wrapper(
            actapi, uri, output_format=output_format, observationTime=observationTime
        ):
            continue

        # Fact: uri -> event
        handle_fact_wrapper(
            actapi.fact("observedIn")
            .source("uri", uri)
            .destination("incident", case_id),
            output_format=output_format,
            observationTime=observationTime,
        )


def handle_argus_event_fqdn(
    actapi: act.api.Act,
    event: defaultdict,
    case_id: Text,
    output_format: Text,
    observationTime: Optional[int] = None,
) -> None:
    "Create fact from fqdn"

    scheme = get_scheme(event)

    # Fact: fqdn -> event
    if event["domain"] and event["domain"]["fqdn"]:
        uri = "{}://{}".format(scheme, event["domain"]["fqdn"])

        # Facts, uri components
        if not handle_uri_wrapper(
            actapi, uri, output_format=output_format, observationTime=observationTime
        ):
            return

        # Fact: uri -> event
        handle_fact_wrapper(
            actapi.fact("observedIn")
            .source("uri", uri)
            .destination("incident", case_id),
            output_format=output_format,
            observationTime=observationTime,
        )


def handle_argus_event(
    actapi: act.api.Act,
    event: Dict,
    content_props: List[Text],
    hash_props: List[Text],
    output_format: Text,
) -> None:
    "handle all facts from an argus event"

    # Use defaultdicts that defaults to None
    properties = defaultdict(lambda: None, event["properties"])
    event = defaultdict(lambda: None, event)

    if not event["associatedCase"]:
        # No incident.
        return

    case_id = "ARGUS-{}".format(event["associatedCase"]["id"])
    subject = event["associatedCase"]["subject"]
    observationTime = event.get("startTimestamp")

    if subject:
        handle_fact_wrapper(
            actapi.fact("name", subject).source("incident", case_id),
            output_format=output_format,
            observationTime=observationTime,
        )

    # For Mitre ATT&CK we use the most specific information we have
    # SubTechnique > Technique > Tactic
    if properties["mitreAttack.subTechniqueId"]:
        handle_argus_event_sub_technique(
            actapi, properties, case_id, output_format, observationTime
        )
    elif properties["mitreAttack.techniqueId"]:
        # If we have technique -> we have an implicit connection to tactic (through technique)
        handle_argus_event_technique(
            actapi, properties, case_id, output_format, observationTime
        )
    elif properties["mitreAttack.tactic"]:
        # If we have tactic (but not technique) we must go through placeholder technique
        handle_argus_event_tactic(
            actapi, properties, case_id, output_format, observationTime
        )

    # Facts: hash/content -> event
    handle_argus_event_hash(
        actapi,
        properties,
        case_id,
        content_props,
        hash_props,
        output_format,
        observationTime,
    )

    # Facts: sourceIP -> event, destinationIP -> event
    handle_argus_event_ip(actapi, event, case_id, output_format, observationTime)

    uris = set()

    if event["uri"]:
        uris.add(refang_uri(event["uri"]))

    if properties["request.uri"]:
        for uri in properties["request.uri"].split("\n"):
            uris.add(refang_uri(uri))

    # Fact: uri -> event. uri can be either in top level field "uri" or property "request.uri"
    for uri in uris:
        if not handle_uri_wrapper(actapi, uri, output_format=output_format):
            continue

        handle_fact_wrapper(
            actapi.fact("observedIn")
            .source("uri", uri)
            .destination("incident", case_id),
            output_format=output_format,
            observationTime=observationTime,
        )

    if not uris:
        # Only construct URI from fqdn if we do not have an uri on the event. The URI
        # Is normally more correct (e.g. scheme and path will be more correctly specified)
        # Facts: uri fqdn -> uri ->  event
        handle_argus_event_fqdn(actapi, event, case_id, output_format, observationTime)



#!/usr/bin/env python3

"""Worker module reading domains and ip-addresses from
stdin, writing result in a format understandable by
generic_uploader.py to stdout"""

import argparse
import socket
import sys
import traceback
import ipaddress
from logging import error, warning
from typing import Any, Dict, Generator, Optional, Text

import requests
import urllib3

import act.api
from act.workers.libs import mnemonic, worker
from act.api.libs import cli

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

IPv4 = "IPv4"
IPv6 = "IPv6"
FQDN = "fqdn"


def parseargs() -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = worker.parseargs("PDNS enrichment")
    parser.add_argument(
        "--pdns-baseurl",
        dest="pdns_baseurl",
        default="https://api.mnemonic.no/",
        help="PassiveDNS API host",
    )
    parser.add_argument(
        "--pdns-timeout", dest="timeout", type=int, default=299, help="Timeout"
    )
    parser.add_argument(
        "--pdns-batch-size",
        dest="pdns_batch_size",
        type=int,
        default=100,
        help="Batch size of pdns queries",
    )
    parser.add_argument("--pdns-apikey", dest="apikey", help="PassiveDNS API key")
    parser.add_argument(
        "--limit",
        dest="limit",
        type=int,
        default=100,
        help="Maximum number of records to return from pdns",
    )

    return parser


def pdns_query(
    pdns_baseurl: str,
    apikey: str,
    query: str,
    timeout: int,
    proxy_string: Optional[Text] = None,
    batch_size: int = 1000,
    limit: int = 0,
) -> Generator[Dict[str, Any], None, None]:
    """Query the passivedns result of an address.
    pdns_baseurl - the url to the passivedns api (https://api.mnemonic.no)
    apikey - PassiveDNS API key with the passivedns role (minimum)
    query - string fqdn or ipv4/6
    timeout - default 299 seconds.
    """

    try:
        qmap = {"baseurl": pdns_baseurl.strip("/"), "query": query}

        pdns_url = "{baseurl}/pdns/v3/{query}".format(**qmap)

        if apikey:
            headers = {"Argus-API-Key": apikey}
        else:
            headers = {}

        yield from mnemonic.batch_query(
            "GET",
            pdns_url,
            headers=headers,
            timeout=timeout,
            proxy_string=proxy_string,
            batch_size=batch_size,
            limit=limit,
        )

    except worker.ResourceLimitExceeded as err:
        error("Resource limits exceeded ({0.__class__.__name__})".format(err))

    except (
        urllib3.exceptions.ReadTimeoutError,
        requests.exceptions.ReadTimeout,
        socket.timeout,
    ) as err:
        warning("Timeout ({0.__class__.__name__}), query: {1}".format(err, query))


def kind(s: Text):
    try:
        ipaddress.IPv4Address(s)
        return IPv4
    except ipaddress.AddressValueError:
        pass
    try:
        ipaddress.IPv6Address(s)
        return IPv6
    except ipaddress.AddressValueError:
        pass
    return FQDN


def process(
    api: act.api.Act,
    pdns_baseurl: str,
    apikey: str,
    timeout: int = 299,
    proxy_string: Optional[Text] = None,
    output_format: Text = "json",
    batch_size: int = 1000,
    limit: int = 0,
) -> None:
    """Read queries from stdin, resolve each one through passivedns
    printing generic_uploader data to stdout"""

    for query in sys.stdin:
        query = query.strip()
        if not query:
            continue

        i = 0
        for row in pdns_query(
            pdns_baseurl,
            apikey,
            timeout=timeout,
            query=query,
            proxy_string=proxy_string,
            batch_size=batch_size,
            limit=limit,
        ):
            rrtype = row["rrtype"]

            i += 1
            if limit == i:
                if kind(query) in (IPv4, IPv6):
                    act.api.helpers.handle_fact(
                        api.fact("excessive", "resolvesTo").source(
                            *act.api.helpers.ip_obj(row["answer"])
                        ),
                        output_format=output_format,
                    )
                else:
                    act.api.helpers.handle_fact(
                        api.fact("excessive", "resolvesTo").source("fqdn", query),
                        output_format=output_format,
                    )

            if rrtype in ("a", "aaaa"):
                act.api.helpers.handle_fact(
                    api.fact("resolvesTo")
                    .source("fqdn", row["query"])
                    .destination(*act.api.helpers.ip_obj(row["answer"])),
                    output_format=output_format,
                )

            elif rrtype == "cname":
                act.api.helpers.handle_fact(
                    api.fact("resolvesTo")
                    .source("fqdn", row["query"])
                    .destination("fqdn", row["answer"]),
                    output_format=output_format,
                )

            elif rrtype == "ptr":
                pass  # We do not insert ptr to act
            else:
                warning("Unsupported rrtype: %s: %s" % (rrtype, row))


def main() -> None:
    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = cli.handle_args(parseargs())
    actapi = worker.init_act(args)

    process(
        actapi,
        args.pdns_baseurl,
        args.apikey,
        args.timeout,
        args.proxy_string,
        args.output_format,
        args.pdns_batch_size,
        args.limit,
    )


def main_log_error() -> None:
    "Main function. Log all exceptions to error"
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()



import argparse
import itertools
import json
import os
import pprint
import re
import sys
import traceback
from collections import defaultdict
from logging import error, info
from typing import Dict, List, Text

import act.api
from act.api.libs import cli

import act
from act.workers.libs import worker


def parseargs() -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = worker.parseargs("Map tool aliases")
    parser.add_argument(
        "--threshold", type=float, default=0.5, help="Threshold for Jaccard index"
    )
    parser.add_argument(
        "--submit", default=False, action="store_true", help="Submit alias to platform"
    )
    parser.add_argument(
        "--exclude_tools",
        default="^\[placeholder\[[a-f0-9]{64}\]\]$",
        help="Tool patterns to exclude",
    )
    return parser


def jaccard(set1: set, set2: set) -> float:
    il = float(len(set1.intersection(set2)))
    ul = float(len(set1.union(set2)))
    return il / ul


def search_tools(api: act.api.Act, ignore_pattern: Text) -> Dict:
    tools = defaultdict(list)
    objects = api.object_search(object_type="tool", limit=10000)
    for obj in objects:
        facts = api.fact_search(
            object_type="content", object_value=obj.value, limit=10000
        )
        for fact in facts:
            if re.match(ignore_pattern, fact.source_object.value):
                continue
            tools[obj.value].append(fact.source_object.value)
    return tools


def get_aliases(api: act.api.Act, tool_names: List) -> Dict:
    aliases = defaultdict(list)
    for tool in tool_names:
        result = api.fact_search(
            object_type="tool", fact_type="alias", object_value=tool
        )
        for fact in result:
            aliases[tool].append(fact.destination_object.value)
    return aliases


def handle_alias(
    api: act.api.Act,
    tool1: Text,
    tool2: Text,
    submit: bool,
    output_format: Text = "json",
):
    try:
        fact = api.fact("alias").bidirectional("tool", tool1, "tool", tool2)
        if submit:
            handle_fact(fact)
        elif output_format == "json":
            print(fact.json())
        else:
            print(fact)
    except act.api.base.ResponseError as e:
        error("Unable to create linked fact: %s" % e)


def process(tools: Dict, aliases: Dict, threshold: float) -> None:
    for tool1, tool2 in itertools.combinations(list(tools.keys()), 2):
        jaccard_index = jaccard(set(tools[tool1]), set(tools[tool2]))
        if jaccard_index > threshold and tool1 not in aliases[tool2]:
            yield tool1, tool2


def main() -> None:
    """Main function"""
    args = cli.handle_args(parseargs())
    actapi = worker.init_act(args)

    if not (args.act_baseurl and args.user_id):
        error("Worker must be configured with --act-baseurl and --userid")
        sys.exit(1)

    tools = search_tools(actapi, args.exclude_tools)
    aliases = get_aliases(actapi, list(tools.keys()))

    for tool1, tool2 in process(tools, aliases, args.threshold):
        handle_alias(actapi, tool1, tool2, args.submit, args.output_format)


def main_log_error() -> None:
    "Main function wrapper. Log all exceptions to error"
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()



#!/usr/bin/env python3

""" Worker for Mitre ATT&CK """

import argparse
import os
import sys
import traceback
from logging import error, info, warning
from typing import Dict, List, Optional, Text, Union

import act.api
from act.api.helpers import Act, handle_fact
from act.api.libs import cli
from act.types.format import ValidationError, format_threat_actor, format_tool
from pyattck import Attck

from act.workers.libs import worker

# This is only for typing
# We can not import these matrices, because then they be loaded
# by pyattack and it fails if you need proxy
AttckMatrice = Union["Enterprise", "PreAttck", "MobileAttck"]

MITRE_TYPES = ["enterprise", "preattack", "mobile"]

DEFAULT_NOTIFY_CACHE = os.path.join(os.environ["HOME"], "act-mitre-attack-notify.cache")


class NotificationError(Exception):
    pass


def parseargs() -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = worker.parseargs("Mitre ATT&CK worker")
    parser.add_argument(
        "--smtphost",
        dest="smtphost",
        help="SMTP host used to send revoked/deprecated objects",
    )
    parser.add_argument(
        "--sender",
        dest="sender",
        help="Sender address used to send revoked/deprecated objects",
    )
    parser.add_argument(
        "--recipient",
        dest="recipient",
        help="Recipient address used to send revoked/deprecated objects",
    )
    parser.add_argument(
        "--type",
        choices=list(MITRE_TYPES),
        help="Specify a single type to download (enterprise, mobile or pre). Default is to fetch all",
    )
    parser.add_argument(
        "--notifycache",
        dest="notifycache",
        help="Cache for revoked/deprecated objects",
        default=DEFAULT_NOTIFY_CACHE,
    )

    return parser


def deprecated_or_revoked(obj):
    """
    Return true if object has a truthy "revoked" or "deprecated" attribute,
    otherwise False
    """
    return getattr(obj, "revoked", None) or getattr(obj, "deprecated", None)


def handle_techniques(
    client: Act,
    technique: "AttckTechnique",
    main_technique: Optional["AttckTechnique"],
    output_format: Text = "json",
) -> List:

    """
    Args:
    client:                Act Client
    technique (str):       Technique or subtechnique ID
    main_technique (str):  If set, technique is a sub technique
    output_format (str):   Fact output if sent to stdout (text | json)
    """

    if deprecated_or_revoked(technique):
        # Object is revoked/deprecated, add to notification list but do not add to facts that should be added to the platform
        return [technique]

    if main_technique:
        handle_fact(
            client.fact("subTechniqueOf")
            .source("technique", technique.id)
            .destination("technique", main_technique.id),
            output_format=output_format,
        )

    handle_fact(
        client.fact("name", technique.name).source("technique", technique.id),
        output_format=output_format,
    )

    # Mitre ATT&CK Tactics are implemented in STIX as kill chain phases with kill_chain_name "mitre-attack"
    for tactic in technique.tactics:
        handle_fact(
            client.fact("accomplishes")
            .source("technique", technique.id)
            .destination("tactic", tactic.id),
            output_format=output_format,
        )

        handle_fact(
            client.fact("name", tactic.name).source("tactic", tactic.id),
            output_format=output_format,
        )

    return []


def add_techniques(
    client: Act, matrice: AttckMatrice, output_format: Text = "json"
) -> List:
    """
        extract objects/facts related to ATT&CK techniques

    Args:
        attack (AttckMatrice):       Attack matrice
        output_format (Text):        "json" or "str" output format

    """

    notify = []

    for technique in matrice.techniques:
        notify += handle_techniques(client, technique, None, output_format)

        for subtechnique in getattr(technique, "subtechniques", []):
            # Pre Attack does not have sub techniques
            notify += handle_techniques(client, subtechnique, technique, output_format)

    return notify


def add_groups(
    client: Act, matrice: AttckMatrice, output_format: Text = "json"
) -> List:
    """
        extract objects/facts related to ATT&CK Threat Actors

    Args:
        attack (AttckMatrice):       Attack matrice
        output_format (Text):        "json" or "str" output format

    """

    notify: List = []

    for actor in matrice.actors:
        if deprecated_or_revoked(actor):
            # Object is revoked, add to notification list but do not add to facts that should be added to the platform
            notify.append(actor)
            continue

        for alias in actor.alias:
            if actor.name != alias:
                handle_fact(
                    client.fact("alias").bidirectional(
                        "threatActor",
                        format_threat_actor(actor.name),
                        "threatActor",
                        format_threat_actor(alias),
                    ),
                    output_format=output_format,
                )

        for tool in actor.known_tools:

            try:
                chain = act.api.fact.fact_chain(
                    client.fact("classifiedAs")
                    .source("content", "*")
                    .destination("tool", format_tool(tool)),
                    client.fact("observedIn")
                    .source("content", "*")
                    .destination("incident", "*"),
                    client.fact("attributedTo")
                    .source("incident", "*")
                    .destination("threatActor", format_threat_actor(actor.name)),
                )

                for fact in chain:
                    handle_fact(fact, output_format=output_format)
            except ValidationError as e:
                error(e)

        for technique in actor.techniques:
            chain = act.api.fact.fact_chain(
                client.fact("observedIn")
                .source("technique", technique.id)
                .destination("incident", "*"),
                client.fact("attributedTo")
                .source("incident", "*")
                .destination("threatActor", format_threat_actor(actor.name)),
            )

            for fact in chain:
                handle_fact(fact, output_format=output_format)

    return notify


def add_software(
    client: Act, matrice: AttckMatrice, output_format: Text = "json"
) -> List:
    """
        extract objects/facts related to ATT&CK Software
        Insert to ACT if client.baseurl is set, if not, print to stdout

    Args:
        attack (AttckMatrice):       Attack matrice
        output_format (Text):        "json" or "str" output format

    """

    notify: List = []

    # Enterprise matrice has malwares and tools, but preattack has none of them
    for software in getattr(matrice, "malwares", []) + getattr(matrice, "tools", []):
        if deprecated_or_revoked(software):
            # Object is revoked/deprecated, add to notification list but do not add to facts that should be added to the platform
            notify.append(software)
            continue

        try:
            tool_name = format_tool(software.name)
        except ValidationError as e:
            error(e)
            continue

        # Tool category
        handle_fact(
            client.fact("category", software.type).source("tool", tool_name),
            output_format=output_format,
        )

        for alias in software.alias:
            try:
                alias_name = format_tool(alias)
            except ValidationError as e:
                error(e)
                continue

            if tool_name != alias_name:
                # Tool category (alias)
                handle_fact(
                    client.fact("category", software.type).source("tool", alias_name),
                    output_format=output_format,
                )
                handle_fact(
                    client.fact("alias").bidirectional(
                        "tool", tool_name, "tool", alias_name
                    ),
                    output_format=output_format,
                )

        for technique in software.techniques:
            try:
                handle_fact(
                    client.fact("implements")
                    .source("tool", format_tool(software.name))
                    .destination("technique", technique.id),
                    output_format=output_format,
                )
            except ValidationError as e:
                error(e)
                continue

    return notify


def notify_cache(filename: str) -> Dict:
    """
    Read notify cache from filename
    Args:
        filename(str):      Cache filename

    """

    cache = {}

    try:
        with open(filename) as f:
            for line in f:
                if line:
                    cache[line.strip()] = True
    except FileNotFoundError:
        warning(
            "Cache file {} not found, will be created if necessary".format(filename)
        )

    return cache


def add_to_cache(filename: str, entry: str) -> None:
    """
    Add entry to cache

    Args:
        filename(str):      Cache filename
        entry(str):         Cache entry
    """

    with open(filename, "a") as f:
        f.write(entry.strip())
        f.write("\n")


def send_notification(
    notify: List,
    smtphost: Text,
    sender: Text,
    recipient: Text,
    model: Text,
) -> List[Text]:
    """
    Process revoked objects

    Args:
        notify(attack[]):   Array of revoked/deprecated objects
        notifycache(str):   Filename of notify cache
        smtphost(str):      SMTP host used to notify of revoked/deprecated objects
        sender(str):        sender address used to notify of revoked/deprecated objects
        recipient(str):     recipient address used to notify of revoked/deprecated objects

    smtphost, sender AND recipient must be set to notify of revoked/deprecated objects

    Return list of IDs that was successfully notified

    """

    notified = []

    if not (smtphost and recipient and sender):
        error(
            "--smtphost, --recipient and --sender must be set to send revoked/deprecated objects on email"
        )
        return []

    body = model + "\n\n"
    warning("[{}]".format(model))

    for obj in notify:
        if getattr(obj, "revoked", None):
            text = "revoked: {}:{}".format(obj.id, obj.name)

        elif getattr(obj, "deprecated", None):
            text = "deprecated: {}:{}".format(obj.id, obj.name)

        else:
            raise NotificationError(
                "object is not deprecated or revoked: {}:{}".format(obj.id, obj.name)
            )

        notified.append(obj.id)

        body += text + "\n"
        warning(text)

    worker.sendmail(
        smtphost,
        sender,
        recipient,
        "Revoked/deprecated objects from MITRE/ATT&CK",
        body,
    )
    info("Email sent to {}".format(recipient))

    return notified


def main() -> None:
    """Main function"""

    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = cli.handle_args(parseargs())

    actapi = worker.init_act(args)

    proxies = (
        {"http": args.proxy_string, "https": args.proxy_string}
        if args.proxy_string
        else None
    )

    attack = Attck(proxies=proxies)

    types = [args.type] if args.type else MITRE_TYPES

    for mitre_type in types:
        if mitre_type not in MITRE_TYPES:
            error(
                "Unknown mitre type: {}. Valid types: {}".format(
                    mitre_type, ",".join(MITRE_TYPES)
                )
            )
            sys.exit(2)

        cache = notify_cache(args.notifycache)

        model = getattr(attack, mitre_type)

        techniques_notify = add_techniques(actapi, model, args.output_format)
        groups_notify = add_groups(actapi, model, args.output_format)
        software_notify = add_software(actapi, model, args.output_format)

        # filter revoked objects from those allready notified
        notify = [
            notify
            for notify in techniques_notify + groups_notify + software_notify
            if notify.id not in cache
        ]

        if notify:
            notified = send_notification(
                notify, args.smtphost, args.sender, args.recipient, mitre_type
            )

            for object_id in notified:
                # Add object to cache, so we will not be notified on the same object on the next run
                add_to_cache(args.notifycache, object_id)


def main_log_error() -> None:
    "Call main() and log all excetions  as errors"
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()



#!/usr/bin/env python3

"""
Worker module fetching ISO 3166 from github to add facts for:
   country -memberOf-> subRegion
   subRegion -memberOf-> region

If --act-baseurl and --userid is specified, add the facts to the platform.
If not, print facts to stdout.
"""


import argparse
import os
import traceback
from logging import error, warning
from typing import Dict, List, Text

import act.api
from act.api.helpers import handle_fact
from act.workers.libs import worker
from act.api.libs import cli


def parseargs() -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = worker.parseargs("Country/region enrichment")
    parser.add_argument(
        "--country-region-url",
        dest="country_region_url",
        default="https://raw.githubusercontent.com/lukes/ISO-3166-Countries-with-Regional-Codes/master/all/all.json",
        help="Country region URL in json format",
    )

    return parser


def process(
    actapi: act.api.Act,
    country_list: List[Dict[str, str]],
    output_format: Text = "json",
) -> None:
    """
    Loop over all ISO-3166 countries and construct facts for
    county -memberOf-> subRegion and subRegion -memberOf-> region.
    """

    for c_map in country_list:
        country_name = c_map["name"]
        sub_region = c_map["sub-region"]
        region = c_map["region"]

        if country_name and sub_region:
            handle_fact(
                actapi.fact("memberOf")
                .source("country", country_name)
                .destination("subRegion", sub_region),
                output_format=output_format,
            )
        else:
            warning("Missing name or sub-region: {}".format(c_map))

        if sub_region and region:
            handle_fact(
                actapi.fact("memberOf")
                .source("subRegion", sub_region)
                .destination("region", region),
                output_format=output_format,
            )
        else:
            warning("Missing sub-region or region: {}".format(c_map))


def main_log_error() -> None:
    "Main function. Log all exceptions to error"
    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = cli.handle_args(parseargs())

    actapi = worker.init_act(args)

    try:
        process(
            actapi,
            worker.fetch_json(
                args.country_region_url, args.proxy_string, args.http_timeout
            ),
            args.output_format,
        )
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()



#!/usr/bin/env python3

"""Fact chain creator helper worker for the ACT platform

Copyright 2021 the ACT project <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
"""

import argparse
import json
import pathlib
import sys
import traceback
from logging import error
from typing import List, Optional, Set, Text, TextIO, Tuple

import act.api
from act.api.helpers import handle_fact
from act.types import types
from act.workers import worker_config
from act.workers.libs import worker
from act.api.libs import cli


class Node:
    """Node in a graph"""

    def __init__(self, name: Text) -> None:
        self.name = name
        self.distance = sys.maxsize
        self.visited = False
        self.neighbours: List[Tuple[Node, int, Text, Text]] = []
        self.previous_direction: Optional[Text] = None
        self.previous_object: Optional[Node] = None
        self.previous_fact: Optional[Text] = None


def path(current_node: Node) -> List[Node]:
    """Backtrack the actuall path after a path is found, rebuilding the
    list of nodes"""

    if not current_node.previous_object:
        return [current_node]

    return path(current_node.previous_object) + [current_node]


def dijkstra(graph: Set[Node], start: Node, stop: Node) -> List[Node]:
    """Compute the shortest path through the types"""

    current: Optional[Node] = start
    start.distance = 0
    unvisited = graph.copy()

    while True:
        if not current:
            return []
        for neighbour, distance, fact, direction in current.neighbours:
            if neighbour.visited:
                continue
            cost = current.distance + distance
            if neighbour.distance >= cost:
                neighbour.distance = cost
                neighbour.previous_object = current
                neighbour.previous_fact = fact
                neighbour.previous_direction = direction

        current.visited = True
        unvisited.remove(current)

        if stop.visited:
            return path(stop)

        current = None
        for node in unvisited:
            if current is None:
                current = node
                continue
            if node.distance < current.distance:
                current = node

        if current and current.distance == sys.maxsize:
            return []


def graph_from_type_def(
    typedef: TextIO, avoids: List[Text], includes: List[Text], avoid_value: int
) -> Set[Node]:
    """Build a graph from a type definition file"""

    typedef.seek(0)
    data = json.load(typedef)

    nodes = dict()
    for fact in data:
        for binding in fact["objectBindings"]:

            if (
                "destinationObjectType" not in binding
                or "sourceObjectType" not in binding
            ):
                continue  # No point in traversing one legged facts

            fact_name = fact["name"]
            dest_objects = binding["destinationObjectType"]
            src_objects = binding["sourceObjectType"]

            if isinstance(dest_objects, str):
                dest_objects = [dest_objects]
            if isinstance(src_objects, str):
                src_objects = [src_objects]

            for dest_object in dest_objects:
                for src_object in src_objects:
                    if dest_object == src_object:
                        continue  # no point in traversing through it self

                    cost = 10

                    if (
                        fact_name in includes
                        or dest_object in includes
                        or src_object in includes
                    ):
                        cost = -100

                    # Avoid overwrites and takes precedence over includes
                    if (
                        fact_name in avoids
                        or dest_object in avoids
                        or src_object in avoids
                    ):
                        cost = 10 * avoid_value

                    if dest_object not in nodes:
                        nodes[dest_object] = Node(dest_object)
                    if src_object not in nodes:
                        nodes[src_object] = Node(src_object)
                    nodes[dest_object].neighbours.append(
                        (nodes[src_object], cost, fact_name, "<")
                    )
                    nodes[src_object].neighbours.append(
                        (nodes[dest_object], cost, fact_name, ">")
                    )

    return set(nodes.values())


def fact_type_value(data: Text) -> Tuple[Text, Text]:
    """Split the argument into fact type and value"""

    typ, val = data.split("/", 1)

    return typ, val


def parseargs() -> argparse.ArgumentParser:
    """Extract command lines argument"""

    parser = worker.parseargs("ACT Fact Chain Helper")
    parser.add_argument(
        "--start",
        type=fact_type_value,
        help="Start type/value (i.e. threatActor/apt21)",
    )
    parser.add_argument(
        "--end", type=fact_type_value, help="End type/value (i.e. tool/zeus)"
    )
    parser.add_argument(
        "--avoid",
        type=worker_config.string_list,
        default=[],
        help="List of fact- or object- types to avoid",
    )
    parser.add_argument(
        "--include",
        type=worker_config.string_list,
        default=[],
        help="List of fact- or object- types to include",
    )
    parser.add_argument(
        "--fact-type-definition",
        type=str,
        default=types.etc_file("fact-types.json"),
        help='fact definition file. default="fact-types.json"',
    )
    parser.add_argument(
        "--avoid-cost",
        type=int,
        default=3,
        help="multiplier used to increase cost of avoided objects or facts. Default=3",
    )

    return parser


def find_start_and_end_nodes(
    graph: Set[Node], start_name: Text, end_name: Text
) -> Tuple[Optional[Node], Optional[Node]]:
    """search through graph looking for nodes with the provded names"""

    start = None
    end = None

    for node in graph:
        if node.name == start_name:
            start = node
        if node.name == end_name:
            end = node

    return start, end


def fact_chain_from_path_result(
    actapi: act.api.Act,
    res: List[Node],
    start: Node,
    end: Node,
    start_value: Text,
    end_value: Text,
) -> List[act.api.fact.FactType]:

    facts: List[act.api.fact.FactType] = []
    prev = None
    value = "*"

    for node in res:
        if prev is None:
            prev = node
            value = start_value if prev.name == start.name else "*"
            continue

        if end.name == node.name:
            value = end_value

        if node.previous_direction == ">":
            if node == end:
                facts.append(
                    actapi.fact(node.previous_fact)
                    .source(prev.name, "*")
                    .destination(node.name, value)
                )
            else:
                facts.append(
                    actapi.fact(node.previous_fact)
                    .source(prev.name, value)
                    .destination(node.name, "*")
                )
        else:
            if node == end:
                facts.append(
                    actapi.fact(node.previous_fact)
                    .destination(prev.name, "*")
                    .source(node.name, value)
                )
            else:
                facts.append(
                    actapi.fact(node.previous_fact)
                    .destination(prev.name, value)
                    .source(node.name, "*")
                )
        prev = node

        value = "*"

    fact_chain: List[act.api.fact.FactType] = act.api.fact.fact_chain(*facts)

    return fact_chain


def main() -> None:
    """main function"""

    # Look for default ini file in "/etc/actworkers.ini" and ~/config/actworkers/actworkers.ini
    # (or replace .config with $XDG_CONFIG_DIR if set)
    args = cli.handle_args(parseargs())

    actapi = worker.init_act(args)

    fact_type_definition_path = (
        pathlib.Path(args.fact_type_definition).expanduser().resolve()
    )

    if not fact_type_definition_path.is_file():
        print(f"{fact_type_definition_path} is not a file.")
        sys.exit(1)

    with fact_type_definition_path.open() as typedef:
        graph = graph_from_type_def(typedef, args.avoid, args.include, args.avoid_cost)

    start_type, start_value = args.start
    end_type, end_value = args.end

    start, end = find_start_and_end_nodes(graph, start_type, end_type)

    if not start:
        print(f"{start_type} is not an object type")
        sys.exit(1)

    if not end:
        print(f"{end_type} is not an object type")
        sys.exit(1)

    res = dijkstra(graph, start, end)

    chain = fact_chain_from_path_result(actapi, res, start, end, start_value, end_value)

    for fact in chain:
        handle_fact(fact, output_format=args.output_format)


def main_log_error() -> None:
    try:
        main()
    except Exception:
        error("Unhandled exception: {}".format(traceback.format_exc()))
        raise


if __name__ == "__main__":
    main_log_error()
